# Copyright (c) 2016 Alex Sherman
# SPDX-License-Identifier: MIT

from deco import concurrent

BODIES = [90]


def run():
    BODIES.append(210)
    simulate()
    simulate.wait()


@concurrent
def simulate():
    print(BODIES)


if __name__ == "__main__":
    run()



# Copyright (c) 2016 Alex Sherman
# SPDX-License-Identifier: MIT

from deco import *
import time


@concurrent.threaded
def threaded_func():
    time.sleep(0.1)


@concurrent
def mp_func():
    time.sleep(0.1)


def sync(func):
    for _ in range(10):
        func()
    func.wait()


if __name__ == "__main__":
    start = time.time()
    sync(mp_func)
    print("Multiprocess duration:", time.time() - start)
    start = time.time()
    sync(threaded_func)
    print("Threaded duration:", time.time() - start)



# Copyright (c) 2016 Alex Sherman
# SPDX-License-Identifier: MIT

from deco import *
import time


@concurrent
def work():
    time.sleep(0.1)


@synchronized
def run():
    for _ in range(100):
        work()


if __name__ == "__main__":
    start = time.time()
    run()
    print("Executing in serial should take 10 seconds")
    print("Executing in parallel took:", time.time() - start, "seconds")



# Copyright (c) 2016 Alex Sherman
# Copyright (c) 2025 Adam Karpierz
# SPDX-License-Identifier: MIT

import deco
import time


@deco.concurrent
def test(sleep_time, i):
    time.sleep(sleep_time)
    return i


@deco.synchronized
def test_size(size):
    results = {}
    for i in range(size):
        results[i] = test(time_duration, i)
    print(results)


SIZE = 100
if __name__ == "__main__":
    processes = [1, 2, 3, 4]
    times = [0, 0.005, 0.01, 0.05, 0.1]  # 0.25]
    for process_count in processes:
        for time_duration in times:
            test.processes = process_count
            test.p = None
            test_size(SIZE)
            exit()
            print(process_count, time_duration)



# Copyright (c) 2016 Alex Sherman
# SPDX-License-Identifier: MIT

import deco
import time
import hashlib


def test_synchronous(sleep_time):
    starttime = time.time()
    m = hashlib.md5(b"Nobody inspects the spammish repetition")
    count = 0
    while((time.time() - starttime) < sleep_time/1000.0):
        m.update(b"p")
        a = m.hexdigest()
        count += 1
    return count


@deco.concurrent
def test(sleep_time):
    starttime = time.time()
    m = hashlib.md5(b"Nobody inspects the spammish repetition")
    count = 0
    while((time.time() - starttime) < sleep_time/1000.0):
        m.update(b" repetition")
        a = m.hexdigest()
        count += 1
    return count


test_time = 2
if __name__ == "__main__":
    times = [0.001, 0.005, 0.01, 0.1, 0.5, 1, 2, 5, 10, 20, 30]
    print("Measuring hashes/sec with single threaded synchronous calls")
    for time_duration in times:
        hashes = 0
        start = time.time()
        iterations = int(test_time/(time_duration/1000.0))
        if iterations > 20000:
            iterations = 20000
        for _ in range(iterations):
            hashes += test_synchronous(time_duration)
        print(time_duration, hashes / (time.time() - start))
    print("Measuring hashes/sec with deco multiprocess calls (3 workers)")
    for time_duration in times:
        hashes = 0
        start = time.time()
        iterations = int(test_time/(time_duration/1000.0))
        if iterations > 20000:
            iterations = 20000
        for _ in range(iterations):
            test(time_duration)
        result = test.wait()
        hashes = sum(result)
        print(hashes / (time.time() - start))



# Copyright (c) 2016 Alex Sherman
# Copyright (c) 2025 Adam Karpierz
# SPDX-License-Identifier: MIT

import unittest
from deco import *


@concurrent
def kwarg_func(kwarg = None):
    kwarg[0] = "kwarged"
    return kwarg


@concurrent
def add_one(value):
    return value + 1


@synchronized
def for_loop(values):
    output = []
    for i in values:
        output.append(add_one(i))
    return [i - 1 for i in output]


class TestCONC(unittest.TestCase):

    def test_kwargs(self):
        list_ = [0]
        kwarg_func(kwarg = list_)
        kwarg_func.wait()
        self.assertEqual(list_[0], "kwarged")

    def test_for_loop(self):
        values = range(30)
        self.assertEqual(list(values), for_loop(values))


if __name__ == "__main__":
    unittest.main(exit=False)



# Copyright (c) 2016 Alex Sherman
# Copyright (c) 2025 Adam Karpierz
# SPDX-License-Identifier: MIT

from multiprocessing import Pool
from multiprocessing.pool import ThreadPool
import inspect
import ast
from . import astutil
import types


def concWrapper(f, args, kwargs):
    result = concurrent.functions[f](*args, **kwargs)
    operations = [inner for outer in args + list(kwargs.values())
                  if type(outer) is argProxy for inner in outer.operations]
    return result, operations


class argProxy(object):

    def __init__(self, arg_id, value):
        self.arg_id = arg_id
        self.operations = []
        self.value = value

    def __getattr__(self, name):
        if name in ["__getstate__", "__setstate__"]:
            raise AttributeError()
        if hasattr(self, 'value') and hasattr(self.value, name):
            return getattr(self.value, name)
        raise AttributeError()

    def __setitem__(self, key, value):
        self.value.__setitem__(key, value)
        self.operations.append((self.arg_id, key, value))

    def __getitem__(self, key):
        return self.value.__getitem__(key)


class synchronized(object):

    def __init__(self, f):
        callerframerecord = inspect.stack()[1][0]
        info = inspect.getframeinfo(callerframerecord)
        self.frame_info = info
        self.orig_f = f
        self.f = None
        self.ast = None
        self.__name__ = f.__name__

    def __get__(self, *args):
        raise NotImplementedError("Decorators from deco cannot be used on class methods")

    def __call__(self, *args, **kwargs):
        if self.f is None:
            source = inspect.getsourcelines(self.orig_f)[0]
            astutil.unindent(source)
            source = "".join(source)
            self.ast = ast.parse(source)
            rewriter = astutil.SchedulerRewriter(concurrent.functions.keys(), self.frame_info)
            rewriter.visit(self.ast.body[0])
            ast.fix_missing_locations(self.ast)
            out = compile(self.ast, "<string>", "exec")
            scope = dict(self.orig_f.__globals__)
            exec(out, scope)
            self.f = scope[self.orig_f.__name__]
        return self.f(*args, **kwargs)


class concurrent(object):

    functions = {}

    @staticmethod
    def custom(constructor = None, apply_async = None):
        @staticmethod
        def _custom_concurrent(*args, **kwargs):
            conc = concurrent(*args, **kwargs)
            if constructor is not None: conc.conc_constructor = constructor
            if apply_async is not None: conc.apply_async = apply_async
            return conc
        return _custom_concurrent

    def __init__(self, *args, **kwargs):
        self.in_progress = False
        self.conc_args = []
        self.conc_kwargs = {}
        if len(args) > 0 and hasattr(args[0], "__call__") and hasattr(args[0], "__name__"):
            self.setFunction(args[0])
        else:
            self.conc_args = args
            self.conc_kwargs = kwargs
        self.results = []
        self.assigns = []
        self.calls = []
        self.arg_proxies = {}
        self.conc_constructor = Pool
        self.apply_async = lambda self, function, args: \
                                  self.concurrency.apply_async(function, args)
        self.concurrency = None

    def __get__(self, *args):
        raise NotImplementedError("Decorators from deco cannot be used on class methods")

    def replaceWithProxies(self, args):
        args_iter = args.items() if type(args) is dict else enumerate(args)
        for i, arg in args_iter:
            if type(arg) is dict or type(arg) is list:
                if not id(arg) in self.arg_proxies:
                    self.arg_proxies[id(arg)] = argProxy(id(arg), arg)
                args[i] = self.arg_proxies[id(arg)]

    def setFunction(self, f):
        concurrent.functions[f.__name__] = f
        self.f_name = f.__name__
        self.__doc__ = f.__doc__
        self.__module__ = f.__module__

    def assign(self, target, *args, **kwargs):
        self.assigns.append((target, self(*args, **kwargs)))

    def call(self, target, *args, **kwargs):
        self.calls.append((target, self(*args, **kwargs)))

    def __call__(self, *args, **kwargs):
        if len(args) > 0 and isinstance(args[0], types.FunctionType):
            self.setFunction(args[0])
            return self
        self.in_progress = True
        if self.concurrency is None:
            self.concurrency = self.conc_constructor(*self.conc_args, **self.conc_kwargs)
        args = list(args)
        self.replaceWithProxies(args)
        self.replaceWithProxies(kwargs)
        result = ConcurrentResult(self.apply_async(self, concWrapper,
                                                   [self.f_name, args, kwargs]))
        self.results.append(result)
        return result

    def apply_operations(self, ops):
        for arg_id, key, value in ops:
            self.arg_proxies[arg_id].value.__setitem__(key, value)

    def wait(self):
        results = []
        while self.results:
            result, operations = self.results.pop().get()
            self.apply_operations(operations)
            results.append(result)
        for assign in self.assigns:
            assign[0][0][assign[0][1]] = assign[1].result()
        self.assigns = []
        for call in self.calls:
            call[0](call[1].result())
        self.calls = []
        self.arg_proxies = {}
        self.in_progress = False
        return results


concurrent.threaded = concurrent.custom(ThreadPool)

class ConcurrentResult(object):

    def __init__(self, async_result):
        self.async_result = async_result

    def get(self):
        return self.async_result.get(3e+6)

    def result(self):
        return self.get()[0]



# Copyright (c) 2016 Alex Sherman
# Copyright (c) 2025 Adam Karpierz
# SPDX-License-Identifier: MIT

from .__about__ import * ; del __about__  # noqa

from . import conc

concurrent = conc.concurrent
synchronized = conc.synchronized



# Copyright (c) 2016 Alex Sherman
# SPDX-License-Identifier: MIT

from deco import *
import time


@concurrent
def work(i):
    time.sleep(0.1)
    return i


@synchronized
def run():
    output = []
    for i in range(100):
        output.append(work(i))
    return output


if __name__ == "__main__":
    start = time.time()
    print(run())
    print("Executing in serial should take 10 seconds")
    print("Executing in parallel took:", time.time() - start, "seconds")



#!/usr/bin/python

# Copyright (c) 2016 Alex Sherman
# SPDX-License-Identifier: MIT

import pygame
import random
import math
import time
from deco import *


iterations = 50


class Body:

    def __init__(self, x, y, vx, vy, mass):
        self.x = x
        self.y = y
        self.vx = vx
        self.vy = vy
        self.mass = mass

    def update(self, fx, fy, dt):
        self.vx += fx / self.mass * dt
        self.vy += fy / self.mass * dt
        self.x += self.vx * dt
        self.y += self.vy * dt

    def distanceSquared(self, other):
        return ((self.x - other.x) ** 2 + (self.y - other.y) ** 2)


@synchronized
def Simulate(body_list, dt):
    next_body_list = {}
    for i in body_list.keys():
        SimulateBody(body_list, next_body_list, i, dt)
    body_list.update(next_body_list)


@concurrent
def SimulateBody(body_list, next_body_list, index, dt):
    simulated_body = body_list[index]
    for _ in range(iterations):
        fx = 0
        fy = 0
        for key in body_list.keys():
            if key == index:
                continue
            body = body_list[key]
            distanceSquared = body.distanceSquared(simulated_body)
            f = body.mass * simulated_body.mass / distanceSquared
            d = distanceSquared ** 0.5
            fx += (body.x - simulated_body.x) / d * f
            fy += (body.y - simulated_body.y) / d * f
        simulated_body.update(fx, fy, dt / iterations)
    next_body_list[index] = simulated_body


if __name__ == "__main__":
    frame_limit = 10
    random.seed(0)

    pygame.init()
    screen_size = 700
    screen = pygame.display.set_mode((screen_size, screen_size))
    points = {}
    point_count = 10
    cs = math.cos(math.pi/3)
    sn = math.sin(math.pi/3)
    key = 0
    for x in range(point_count):
        for y in range(point_count):
            dx = point_count / 2 - x
            dy = point_count / 2 - y
            d_len = math.pow(math.pow(dx, 2) + math.pow(dy, 2), .5)
            if d_len == 0:
                m = 1000
                vx = 0
                vy = 0
            else:
                dx /= d_len
                dy /= d_len
                m = 10000 * random.random()
                vx = (dx * cs - dy * sn) * 30 + (random.random() - .5) / 700
                vy = (dx * sn + dy * cs) * 30 + (random.random() - .5) / 700

            points[key] = Body(1.0 * screen_size * x / point_count,
                               1.0 * screen_size * y / point_count,
                               vx, vy, m)
            key += 1

    body_count = len(points)

    i = 0
    start = time.time()
    while True:
        screen.fill((0, 0, 0))
        for point in points.values():
            pygame.draw.circle(
                screen, (128, 128, 128),
                (int(point.x), int(point.y)),
                int(math.pow(point.mass/200, .5)), 0)

        pygame.display.update()
        Simulate(points, .1)
        if i < frame_limit:
            i += 1
            if i == frame_limit:
                print("Time:", time.time() - start)

        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                pygame.quit()
                exit()



# Copyright (c) 2016 Alex Sherman
# Copyright (c) 2025 Adam Karpierz
# SPDX-License-Identifier: MIT

import unittest
import ast
import inspect
import time
from deco import *


@concurrent
def conc_func(*args, **kwargs):
    time.sleep(0.1)
    return kwargs


@synchronized
def body_cases():
    conc_func()
    a = True if False else False
    b = (lambda : True)()


@synchronized
def tainted_return():
    data = []
    data.append(conc_func(data))
    return data


@synchronized
def len_of_append():
    data = []
    data.append(conc_func([]))
    derp = len(data)
    return derp


def indented():
    @synchronized
    def _indented():
        conc_func()

    return _indented()


@synchronized
def kwarged_sync(**kwargs):
    data = []
    data.append(conc_func(**kwargs))
    return data[0]


@synchronized
def subscript_args():
    d = type('', (object,), {"items": {(0,0): 0}})()
    conc_func(d.items[0, 0])
    # Read d to force a synchronization event
    d = d
    output = conc_func.in_progress
    return output


@synchronized
def list_comp():
    result = [conc_func(i = i) for i in range(10)]
    return result


class TestAST(unittest.TestCase):

    # This just shouldn't throw any exceptions
    def test_body_cases(self):
        body_cases()

    # This just shouldn't throw any exceptions
    def test_indent_cases(self):
        indented()

    # This just shouldn't throw any exceptions
    def test_tainted_return(self):
        tainted_return()

    def test_subscript_args(self):
        self.assertFalse(subscript_args())

    def test_kwarged_sync(self):
        self.assertTrue(kwarged_sync(test = "test")["test"] == "test")

    def test_wait_after_append(self):
        self.assertEqual(len_of_append(), 1)

    def test_list_comp(self):
        self.assertEqual(list_comp(),[{'i': i} for i in range(10)])


if __name__ == "__main__":
    unittest.main(exit=False)



# Copyright (c) 2025 Adam Karpierz
# SPDX-License-Identifier: MIT

import unittest
import sys
import logging

from . import test_dir

log = logging.getLogger(__name__)


def test_suite(names=None, omit=()):
    from . import __name__ as pkg_name
    from . import __path__ as pkg_path
    import unittest
    import pkgutil
    if names is None:
        names = [name for _, name, _ in pkgutil.iter_modules(pkg_path)
                 if name.startswith("test_") and name not in omit]
    names = [".".join((pkg_name, name)) for name in names]
    tests = unittest.defaultTestLoader.loadTestsFromNames(names)
    return tests


def main(argv=sys.argv[1:]):
    import runpy
    print("Running tests\n", file=sys.stderr)
    tests = test_suite(argv or None)
    # result = unittest.TextTestRunner(verbosity=2).run(tests)
    # return 0 if result.wasSuccessful() else 1
    result1 = runpy.run_path(str(test_dir/"test_ast.py"),  run_name="__main__")
    result2 = runpy.run_path(str(test_dir/"test_conc.py"), run_name="__main__")
    result3 = runpy.run_path(str(test_dir/"tman_conc.py"), run_name="__main__")
    return 0 if result1 and result2 and result3 else 1


if __name__.rpartition(".")[-1] == "__main__":
    # logging.basicConfig(level=logging.INFO)
    # logging.basicConfig(level=logging.DEBUG)
    sys.exit(main())



# Copyright (c) 2016 Alex Sherman
# SPDX-License-Identifier: MIT

from deco import *
import time
import random
from collections import defaultdict


@concurrent  # We add this for the concurrent function
def process_lat_lon(lat, lon, data):
    time.sleep(0.1)
    return data[lat + lon]


@synchronized  # And we add this for the function which calls the concurrent function
def process_data_set(data):
    results = defaultdict(dict)
    for lat in range(5):
        for lon in range(5):
            results[lat][lon] = process_lat_lon(lat, lon, data)
    return dict(results)


if __name__ == "__main__":
    random.seed(0)
    data = [random.random() for _ in range(200)]
    start = time.time()
    print(process_data_set(data))
    print(time.time() - start)



# Copyright (c) 2016 Alex Sherman
# Copyright (c) 2025 Adam Karpierz
# SPDX-License-Identifier: MIT

import ast
from ast import NodeTransformer, copy_location


def unindent(source_lines):
    for i, line in enumerate(source_lines):
        source_lines[i] = line.lstrip()
        if source_lines[i][:3] == "def":
            break

def Call(func, args=None, kwargs=None):
    if args is None:
        args = []
    if kwargs is None:
        kwargs = []
    return ast.Call(func, args, kwargs)


class SchedulerRewriter(NodeTransformer):

    def __init__(self, concurrent_funcs, frameinfo):
        self.arguments = set()
        self.concurrent_funcs = concurrent_funcs
        self.encountered_funcs = set()
        self.line_offset = frameinfo.lineno - 1
        self.filename = frameinfo.filename

    def references_arg(self, node):
        if not isinstance(node, ast.AST):
            return False
        if type(node) is ast.Name:
            return type(node.ctx) is ast.Load and node.id in self.arguments
        for field in node._fields:
            if field == "body": continue
            value = getattr(node, field)
            if not hasattr(value, "__iter__"):
                value = [value]
            if any([self.references_arg(child) for child in value]):
                return True
        return False

    def not_implemented_error(self, node, message):
        return NotImplementedError(f"{self.filename}({node.lineno + self.line_offset}) {message}")

    @staticmethod
    def top_level_name(node):
        if type(node) is ast.Name:
            return node.id
        elif type(node) is ast.Subscript or type(node) is ast.Attribute:
            return SchedulerRewriter.top_level_name(node.value)
        return None

    def is_concurrent_call(self, node):
        return (type(node) is ast.Call
                and type(node.func) is ast.Name
                and node.func.id in self.concurrent_funcs)

    def encounter_call(self, call):
        self.encountered_funcs.add(call.func.id)
        for arg in call.args:
            arg_name = SchedulerRewriter.top_level_name(arg)
            if arg_name is not None:
                self.arguments.add(arg_name)

    def get_waits(self):
        return [ast.Expr(Call(ast.Attribute(ast.Name(fname,  ast.Load()),
                                                     'wait', ast.Load())))
                for fname in self.encountered_funcs]

    def visit_Call(self, node):
        if self.is_concurrent_call(node):
            raise self.not_implemented_error(node, "The usage of the @concurrent function "
                                                   "is unsupported")
        node = self.generic_visit(node)
        return node

    def generic_visit(self, node):
        if ((isinstance(node, ast.stmt)
             and self.references_arg(node)) or isinstance(node, ast.Return)):
            return self.get_waits() + [node]
        return NodeTransformer.generic_visit(self, node)

    def makeCall(self, func, args = [], keywords = []):
        return ast.Call(func = func, args = args, keywords = keywords)

    def makeLambda(self, args, call):
        return ast.Lambda(ast.arguments(posonlyargs = [], args = args, defaults = [],
                                        kwonlyargs = [], kw_defaults = []), call)

    def visit_Expr(self, node):
        if type(node.value) is ast.Call:
            call = node.value
            if self.is_concurrent_call(call):
                self.encounter_call(call)
                return node
            elif any([self.is_concurrent_call(arg) for arg in call.args]):
                conc_args = [(i, arg) for i, arg in enumerate(call.args)
                             if self.is_concurrent_call(arg)]
                if len(conc_args) > 1:
                    raise self.not_implemented_error(call, "Functions with multiple @concurrent "
                                                           "parameters are unsupported")
                conc_call = conc_args[0][1]
                if isinstance(call.func, ast.Attribute):
                    self.arguments.add(SchedulerRewriter.top_level_name(call.func.value))
                self.encounter_call(conc_call)
                call.args[conc_args[0][0]] = ast.Name("__value__", ast.Load())
                args = [ast.arg("__value__", None)]
                call_lambda = self.makeLambda(args, call)
                copy_location_kwargs = {
                    "func": ast.Attribute(conc_call.func, 'call', ast.Load()),
                    "args": [call_lambda] + conc_call.args,
                    "keywords": conc_call.keywords
                }
                return copy_location(ast.Expr(ast.Call(**copy_location_kwargs)), node)
        return self.generic_visit(node)

    # List comprehensions are self contained, so no need to add to self.arguments
    def visit_ListComp(self, node):
        if self.is_concurrent_call(node.elt):
            self.encounter_call(node.elt)
            wrapper = self.makeCall(
                func = ast.Name('list', ast.Load()),
                args = [
                    self.makeCall(
                        func = ast.Name('map', ast.Load()),
                        args = [
                            self.makeLambda(
                                [ast.arg(arg='r')],
                                self.makeCall(func = ast.Attribute(ast.Name('r', ast.Load()),
                                                                   'result', ast.Load()))
                            ),
                            node
                        ]
                    )
                ]
            )
            return wrapper
        return self.generic_visit(node)

    def is_valid_assignment(self, node):
        if not (type(node) is ast.Assign and self.is_concurrent_call(node.value)):
            return False
        if len(node.targets) != 1:
            raise self.not_implemented_error(node, "Concurrent assignment does not support "
                                                   "multiple assignment targets")
        if not type(node.targets[0]) is ast.Subscript:
            raise self.not_implemented_error(node, "Concurrent assignment only implemented "
                                                   "for index based objects")
        return True

    def visit_Assign(self, node):
        if self.is_valid_assignment(node):
            call = node.value
            self.encounter_call(call)
            name = node.targets[0].value
            self.arguments.add(SchedulerRewriter.top_level_name(name))
            # Check ast.slice compatibility
            index = node.targets[0].slice
            if hasattr(index, "value"):
                # For Python <= 3.8
                index = index.value
            call.func = ast.Attribute(call.func, 'assign', ast.Load())
            call.args = [ast.Tuple([name, index], ast.Load())] + call.args
            return copy_location(ast.Expr(call), node)
        return self.generic_visit(node)

    def visit_FunctionDef(self, node):
        node.decorator_list = []
        node = self.generic_visit(node)
        node.body += self.get_waits()
        return node
